from datetime import datetime

from petisco.base.misc.singleton import Singleton


class ApplicationInfo(metaclass=Singleton):
    name: str
    organization: str
    version: str
    deployed_at: datetime

    def __init__(self, **kwargs):
        self.name = kwargs.get("name")
        self.organization = kwargs.get("organization")
        self.version = kwargs.get("version")
        self.deployed_at = kwargs.get("deployed_at")



from typing import List, Type

from petisco.base.domain.message.message_subscriber import MessageSubscriber
from petisco.extra.rabbitmq.application.message.formatter.rabbitmq_message_subscriber_queue_name_formatter import (
    RabbitMqMessageSubscriberQueueNameFormatter,
)
from petisco.extra.rabbitmq.shared.queue_config import QueueConfig
from petisco.extra.rabbitmq.shared.rabbitmq_connector import RabbitMqConnector
from petisco.extra.rabbitmq.shared.rabbitmq_declarer import RabbitMqDeclarer
from petisco.extra.rabbitmq.shared.rabbitmq_exchange_name_formatter import (
    RabbitMqExchangeNameFormatter,
)


class RabbitMqMessageSubcribersConfigurer:
    def __init__(
        self,
        organization: str,
        service: str,
        connector: RabbitMqConnector,
        queue_config: QueueConfig,
    ):
        self._connector = connector
        self._exchange_name = f"{organization}.{service}"
        self._retry_exchange_name = RabbitMqExchangeNameFormatter.retry(
            self._exchange_name
        )
        self._dead_letter_exchange_name = RabbitMqExchangeNameFormatter.dead_letter(
            self._exchange_name
        )
        self.rabbitmq = RabbitMqDeclarer(
            connector=self._connector, channel_name=self._exchange_name
        )
        self._configured_subscribers = []
        self.queue_config = queue_config

    def execute(self, subscribers):
        self._configure_exchanges()
        self._declare_queues(
            self._exchange_name,
            self._retry_exchange_name,
            self._dead_letter_exchange_name,
            subscribers,
        )
        self._configured_subscribers += subscribers

    def clear_subscribers(self, subscribers):
        self._configured_subscribers += subscribers
        self.clear()

    def clear(self):
        self._delete_exchange()
        self._delete_queues()

    def _configure_exchanges(self):
        self.rabbitmq.declare_exchange(self._exchange_name)
        self.rabbitmq.declare_exchange(self._retry_exchange_name)
        self.rabbitmq.declare_exchange(self._dead_letter_exchange_name)

    def _delete_exchange(self):
        self.rabbitmq.delete_exchange(self._exchange_name)
        self.rabbitmq.delete_exchange(self._retry_exchange_name)
        self.rabbitmq.delete_exchange(self._dead_letter_exchange_name)

    def _delete_queues(self):
        for SubscriberClass in self._configured_subscribers:

            subscriber = SubscriberClass()
            for subscriber_info in subscriber.get_message_subscribers_info():

                if subscriber_info.message_type == "message":
                    # if subscriber_info is subscribed to message it will be consuming from store queue
                    break

                queue_name = RabbitMqMessageSubscriberQueueNameFormatter.format(
                    subscriber_info, exchange_name=self._exchange_name
                )
                retry_queue_name = (
                    RabbitMqMessageSubscriberQueueNameFormatter.format_retry(
                        subscriber_info, exchange_name=self._exchange_name
                    )
                )
                dead_letter_queue_name = (
                    RabbitMqMessageSubscriberQueueNameFormatter.format_dead_letter(
                        subscriber_info, exchange_name=self._exchange_name
                    )
                )

                suffix = subscriber.get_subscriber_name()
                name = f"{queue_name}.{suffix}"
                retry_name = f"{retry_queue_name}.{suffix}"
                dead_letter_name = f"{dead_letter_queue_name}.{suffix}"
                self.rabbitmq.delete_queue(name)
                self.rabbitmq.delete_queue(retry_name)
                self.rabbitmq.delete_queue(dead_letter_name)

    def _declare_queues(
        self,
        exchange_name: str,
        retry_exchange_name: str,
        dead_letter_exchange_name: str,
        subscribers: List[Type[MessageSubscriber]],
    ):

        for SubscriberClass in subscribers:
            subscriber = SubscriberClass()
            for subscriber_info in subscriber.get_message_subscribers_info():
                if subscriber_info.message_type == "message":
                    # if subscriber_info is subscribed to message it will be consuming from store queue
                    break

                base_queue_name = RabbitMqMessageSubscriberQueueNameFormatter.format(
                    subscriber_info, exchange_name=exchange_name
                )
                base_retry_queue_name = (
                    RabbitMqMessageSubscriberQueueNameFormatter.format_retry(
                        subscriber_info, exchange_name=exchange_name
                    )
                )
                base_dead_letter_queue_name = (
                    RabbitMqMessageSubscriberQueueNameFormatter.format_dead_letter(
                        subscriber_info, exchange_name=exchange_name
                    )
                )
                routing_key = base_queue_name
                suffix = subscriber.get_subscriber_name()

                queue_name = f"{base_queue_name}.{suffix}"
                retry_queue_name = f"{base_retry_queue_name}.{suffix}"
                dead_letter_queue_name = f"{base_dead_letter_queue_name}.{suffix}"

                main_ttl = self.queue_config.get_main_ttl(queue_name)
                retry_ttl = self.queue_config.get_retry_ttl(queue_name)

                self.rabbitmq.declare_queue(
                    queue_name=queue_name,
                    dead_letter_exchange=f"dead_letter.{exchange_name}",
                    dead_letter_routing_key=f"dead_letter.{queue_name}",
                    message_ttl=main_ttl,
                )
                self.rabbitmq.declare_queue(
                    queue_name=retry_queue_name,
                    dead_letter_exchange=exchange_name,
                    dead_letter_routing_key=f"retry.{queue_name}",
                    message_ttl=retry_ttl,
                )
                self.rabbitmq.declare_queue(queue_name=dead_letter_queue_name)

                self.rabbitmq.bind_queue(
                    exchange_name=exchange_name,
                    queue_name=queue_name,
                    routing_key=routing_key,
                )
                self.rabbitmq.bind_queue(
                    exchange_name=exchange_name,
                    queue_name=queue_name,
                    routing_key=f"retry.{queue_name}",
                )
                self.rabbitmq.bind_queue(
                    exchange_name=retry_exchange_name,
                    queue_name=retry_queue_name,
                    routing_key=f"retry.{queue_name}",
                )
                self.rabbitmq.bind_queue(
                    exchange_name=dead_letter_exchange_name,
                    queue_name=dead_letter_queue_name,
                    routing_key=f"dead_letter.{queue_name}",
                )



from typing import Dict, List


def create_text_meta(meta: Dict) -> List[Dict]:
    text = None
    if meta:
        text = ""
        for key, value in meta.items():
            text += f"\n>{key}: {value}"
    return text



import os
from typing import Dict

from petisco.base.misc.builder import Builder


class Dependency:
    name: str
    default_builder: Builder
    envar_modifier: str = None
    builders: Dict[str, Builder] = None

    def __init__(
        self,
        name: str,
        default_builder: Builder,
        envar_modifier: str = None,
        builders: Dict[str, Builder] = None,
    ):
        self.name = name
        self.default_builder = default_builder
        self.envar_modifier = envar_modifier
        self.builders = builders

    def get_instance(self):
        if not self.envar_modifier:
            return self.default_builder.build()

        modifier = os.getenv(self.envar_modifier)
        if not modifier or not self.builders or modifier not in self.builders:
            return self.default_builder.build()
        else:
            instance = self.builders.get(modifier).build()
            return instance



from datetime import datetime, timedelta
from typing import Callable, List, Optional

from pydantic import BaseSettings, Field

from petisco.base.application.application_configurer import ApplicationConfigurer
from petisco.base.application.application_info import ApplicationInfo
from petisco.base.application.dependency_injection.container import Container
from petisco.base.application.dependency_injection.dependency import Dependency
from petisco.base.application.notifier.notifier import Notifier
from petisco.base.application.notifier.notifier_message import NotifierMessage
from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.message.domain_event_bus import DomainEventBus
from petisco.extra.rabbitmq.dependencies import get_default_message_dependencies
from petisco.extra.slack.dependencies import get_default_notifier_dependencies


class Application(BaseSettings):
    name: str
    version: str
    organization: str
    deployed_at: Optional[datetime]
    environment: str = Field("local", env="ENVIRONMENT")
    dependencies_provider: Optional[Callable[..., List[Dependency]]] = lambda: []
    configurers: Optional[List[ApplicationConfigurer]] = []

    def __init__(self, **data) -> None:
        ApplicationInfo(
            name=data["name"],
            organization=data["organization"],
            version=data["version"],
            deployed_at=data.get("deployed_at"),
        )
        super().__init__(**data)

    def configure(self, testing: bool = False):

        before_dependecies_configurers = [
            configurer
            for configurer in self.configurers
            if configurer.execute_after_dependencies is False
        ]
        for configurer in before_dependecies_configurers:
            configurer.execute(testing)

        Container.set_dependencies(self.get_dependencies())

        after_dependencies_configurers = [
            configurer
            for configurer in self.configurers
            if configurer.execute_after_dependencies is True
        ]
        for configurer in after_dependencies_configurers:
            configurer.execute(testing)

    def get_dependencies(self) -> List[Dependency]:
        default_dependencies = (
            get_default_message_dependencies() + get_default_notifier_dependencies()
        )
        default_dependencies_dict = {
            dependency.name: dependency for dependency in default_dependencies
        }
        provided_dependencies = self.dependencies_provider()
        given_dependencies_dict = {
            dependency.name: dependency for dependency in provided_dependencies
        }
        merged_dependecies = {**default_dependencies_dict, **given_dependencies_dict}
        return list(merged_dependecies.values())

    def clear(self):
        Container.clear()

    def info(self):
        info = self.dict()
        info["deployed_at"] = (
            self.deployed_at.strftime("%m/%d/%Y, %H:%M:%S")
            if self.deployed_at
            else None
        )
        info["dependencies"] = {
            dependency.name: dependency.get_instance().info()
            for dependency in self.get_dependencies()
        }
        del info["dependencies_provider"]
        del info["configurers"]
        return info

    def was_deploy_few_minutes_ago(self, minutes: int = 25):
        return datetime.utcnow() < self.deployed_at + timedelta(minutes=minutes)

    def publish_domain_event(self, domain_event: DomainEvent):
        try:
            domain_event_bus: DomainEventBus = Container.get("domain_event_bus")
            domain_event_bus.publish(domain_event)
        except:  # noqa
            raise TypeError(
                'To publish an event to the domain event bus, please add a dependency with name "domain_event_bus" on Application dependencies'
            )

    def notify(self, message: NotifierMessage):
        try:
            notifier: Notifier = Container.get("notifier")
            notifier.publish(message)
        except:  # noqa
            raise TypeError(
                'To notify the deploy, please add a dependency with name "notifier" on Application dependencies'
            )



from typing import List

from petisco.base.domain.message.message_configurer import MessageConfigurer
from petisco.base.domain.message.message_subscriber import MessageSubscriber
from petisco.extra.rabbitmq.application.message.configurer.rabbitmq_message_store_configurer import (
    RabbitMqMessageStoreConfigurer,
)
from petisco.extra.rabbitmq.application.message.configurer.rabbitmq_message_subscribers_configurer import (
    RabbitMqMessageSubcribersConfigurer,
)
from petisco.extra.rabbitmq.shared.queue_config import QueueConfig
from petisco.extra.rabbitmq.shared.rabbitmq_connector import RabbitMqConnector


class RabbitMqMessageConfigurer(MessageConfigurer):
    def __init__(
        self,
        organization: str,
        service: str,
        connector: RabbitMqConnector = RabbitMqConnector(),
        queue_config: QueueConfig = QueueConfig.default(),
        use_store_queues: bool = True,
    ):
        self._use_store_queues = use_store_queues
        self.subscribers_configurer = RabbitMqMessageSubcribersConfigurer(
            organization, service, connector, queue_config
        )
        self.store_configurer = RabbitMqMessageStoreConfigurer(
            organization, service, connector, queue_config
        )

    def configure(self):
        self.configure_subscribers([])

    def configure_subscribers(
        self,
        subscribers: List[MessageSubscriber],
        clear_subscriber_before: bool = False,
        clear_store_before: bool = False,
    ):
        if subscribers is None:
            subscribers = []

        if clear_subscriber_before:
            self.subscribers_configurer.clear_subscribers(subscribers)
        if clear_store_before and self._use_store_queues:
            self.store_configurer.clear()

        self.subscribers_configurer.execute(subscribers)
        if self._use_store_queues:
            self.store_configurer.execute()

    def clear_subscribers(self, subscribers: List[MessageSubscriber]):
        if subscribers is None:
            subscribers = []

        self.subscribers_configurer.clear_subscribers(subscribers)
        if self._use_store_queues:
            self.store_configurer.clear()

    def clear(self):
        self.subscribers_configurer.clear()
        if self._use_store_queues:
            self.store_configurer.clear()



import elasticapm
from fastapi import HTTPException
from loguru import logger
from meiga import Result

from petisco import DEFAULT_HTTP_ERROR_MAP
from petisco.base.application.controller.error_map import ErrorMap
from petisco.base.application.controller.http_error import (
    DEFAULT_HTTP_ERROR_DETAIL,
    HttpError,
)
from petisco.base.domain.errors.domain_error import DomainError
from petisco.base.domain.errors.unknown_error import UnknownError


def fastapi_failure_handler(result: Result, error_map: ErrorMap):
    domain_error = result.value
    error_type = type(domain_error)
    http_error = error_map.get(error_type, HttpError())

    elasticapm.set_custom_context({"http_response": str(http_error)})

    internal_error_message = None
    if isinstance(result.value, UnknownError):
        internal_error_message = str(result.value.__dict__)
    elif error_type not in error_map:
        internal_error_message = (
            f"Error '{result.value.__class__.__name__}' is not mapped in controller"
        )

    if internal_error_message is not None:
        logger.error(internal_error_message)
        elasticapm.set_custom_context(
            {"internal_error_message": internal_error_message}
        )

    detail = http_error.detail
    if isinstance(domain_error, DomainError):
        detail = (
            http_error.detail
            if http_error.detail != DEFAULT_HTTP_ERROR_DETAIL
            else domain_error.detail()
        )
    else:
        logger.warning(
            f"fastapi_failure_handler: {domain_error.__class__.__name__} is not a DomainError (warning "
            f"mapping error)"
        )

    http_exception = HTTPException(
        status_code=http_error.status_code, detail=detail, headers=http_error.headers
    )

    if (
        isinstance(domain_error, DomainError)
        and type(domain_error) not in DEFAULT_HTTP_ERROR_MAP.keys()
    ):
        logger.error(f"DomainError:  {domain_error.__repr__()}")
        logger.error(f"HTTPException: {http_exception.__repr__()}")

    raise http_exception



from abc import abstractmethod

from petisco.base.misc.interface import Interface


class MessageChaos(Interface):
    @abstractmethod
    def nack_simulation(self, *args, **kwargs) -> bool:
        raise NotImplementedError

    @abstractmethod
    def failure_simulation(self, *args, **kwargs) -> bool:
        raise NotImplementedError

    @abstractmethod
    def delay(self):
        raise NotImplementedError



from abc import abstractmethod
from typing import NoReturn

from petisco.base.misc.interface import Interface


class ApplicationConfigurer(Interface):
    def __init__(self, execute_after_dependencies: bool = False):
        self.execute_after_dependencies = execute_after_dependencies

    @abstractmethod
    def execute(self, testing: bool = False) -> NoReturn:
        raise NotImplementedError()



from typing import Any


class Builder:
    def __init__(self, klass: Any, is_builder: bool = False, *args, **kwargs):
        self.klass = klass
        self.args = args
        self.kwargs = kwargs
        self.is_builder = is_builder

    def build(self):
        try:
            if self.is_builder:
                instance = self.klass.build(*self.args, **self.kwargs)
            else:
                instance = self.klass(*self.args, **self.kwargs)
        except Exception as exc:
            raise RuntimeError(
                f"Error instantiating {self.klass.__name__}\n{repr(exc)}"
            )

        return instance



from typing import Dict

from petisco.base.application.notifier.notifier_message import NotifierMessage
from petisco.extra.slack.application.notifier.create_text_meta import create_text_meta
from petisco.extra.slack.application.notifier.slack_notifier_message_converter import (
    SlackNotifierMessageConverter,
)


class BlocksSlackNotifierMessageConverter(SlackNotifierMessageConverter):
    def __init__(self, slack_accessory: Dict = None):
        self.slack_accessory = slack_accessory

    def convert(self, notifier_message: NotifierMessage):
        header_block = self._create_header_block(notifier_message.title)
        message_block = self._create_message_block(notifier_message)
        divider_block = {"type": "divider"}
        return [header_block, message_block, divider_block]

    def _create_header_block(self, title: str):
        return {
            "type": "header",
            "text": {
                "type": "plain_text",
                "text": title,
                "emoji": True,
            },
        }

    def _create_message_block(self, notifier_message: NotifierMessage):
        text_message = ""
        if notifier_message.message:
            text_message += f"{notifier_message.message}"

        text_meta = create_text_meta(notifier_message.meta)
        if text_meta:
            text_message += text_meta
        message_block = {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": text_message,
            },
        }

        if self.slack_accessory:
            message_block["accessory"] = self.slack_accessory

        return message_block



from abc import abstractmethod
from typing import Dict, List, Union

from petisco.base.application.notifier.notifier_exception_message import (
    NotifierExceptionMessage,
)
from petisco.base.application.notifier.notifier_message import NotifierMessage
from petisco.base.misc.interface import Interface


class SlackNotifierMessageConverter(Interface):
    @abstractmethod
    def convert(
        self, notifier_message: Union[NotifierMessage, NotifierExceptionMessage]
    ) -> List[Dict]:
        """
        Should return a list of blocks. See: https://api.slack.com/reference/block-kit/blocks
        """
        raise NotImplementedError



from abc import abstractmethod
from typing import Callable, List, Type

from petisco.base.domain.message.message_subscriber import MessageSubscriber
from petisco.base.misc.interface import Interface


class MessageConsumer(Interface):
    @classmethod
    def __repr__(cls):
        return cls.__name__

    @abstractmethod
    def add_subscribers(self, subscribers: List[MessageSubscriber]):
        raise NotImplementedError

    @abstractmethod
    def add_subscriber_on_dead_letter(self, subscriber: Type[MessageSubscriber]):
        raise NotImplementedError

    @abstractmethod
    def add_subscriber_on_queue(self, queue_name: str, handler: Callable):
        raise NotImplementedError

    @abstractmethod
    def unsubscribe_subscriber_on_queue(self, queue_name: str):
        raise NotImplementedError

    @abstractmethod
    def resume_subscriber_on_queue(self, queue_name: str):
        raise NotImplementedError

    @abstractmethod
    def start(self):
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        raise NotImplementedError



import inspect
from dataclasses import dataclass
from typing import Callable, List

from loguru import logger

from petisco.base.domain.persistence.interface_database import Database
from petisco.base.misc.singleton import Singleton


@dataclass
class Persistence(metaclass=Singleton):
    def __init__(self):
        self._databases = {}

    def __repr__(self):
        return f"Persistence: {str(self.get_info())}"

    def get_info(self):
        return {name: database.info() for name, database in self._databases.items()}

    @staticmethod
    def info():
        return Persistence.get_instance().get_info()

    @staticmethod
    def get_instance():
        try:
            return Persistence()
        except Exception as e:  # noqa E722
            frame_info = inspect.stack()[1]
            raise ImportError(
                f"Persistence must be configured. If not, you cannot obtain models\n"
                f"Following code must be executed after Persistence initialization:\n"
                f"\tfilename: {frame_info.filename}\n"
                f"\tlineno: {frame_info.lineno}\n"
                f"\tfunction: {frame_info.function}\n"
                f"\tcode_context: {frame_info.code_context}\n\n"
            )

    def add(self, database: Database, skip_if_exist: bool = False):
        if database.name in self._databases:
            if skip_if_exist is False:
                raise NameError(
                    f"Database {database.name} is already added to Persistence"
                )
        else:
            self._databases[database.name] = database

    def remove(self, database_name: str, skip_if_not_exist: bool = False):
        if database_name in self._databases:
            self._databases[database_name].delete()
            del self._databases[database_name]
        else:
            if skip_if_not_exist is False:
                raise IndexError(
                    f"Database cannot be removed. {database_name} not exists"
                )

    def create(self):
        for database in self._databases.values():
            database.create()

    def delete(self):
        for database in self._databases.values():
            database.delete()

    def clear_data(self, database_name: str = None):
        databases = self._databases
        if database_name is not None:
            if database_name not in self._databases:
                raise IndexError(
                    f"Database cannot clear the data. {database_name} not exists"
                )
            databases = [self._databases.get(database_name)]
        for database in databases.values():
            database.clear_data()

    @staticmethod
    def exist():
        databases = Persistence.get_instance()._databases
        if len(databases) < 1:
            return False
        else:
            return True

    @staticmethod
    def is_available(database_name: str = None):
        def log_warning(message: str):
            logger.debug(message)

        databases = Persistence.get_instance()._databases
        if database_name is not None:
            if database_name not in databases:
                raise IndexError(
                    f"Database cannot return is_available. {database_name} not exists"
                )
            databases = {database_name: databases.get(database_name)}
        if len(databases) < 1:
            log_warning("Persistence databases are empty")
            return False
        for database_name, database in databases.items():
            if not database.is_available():
                log_warning(f"Database {database_name} is not available")
                return False
        return True

    @staticmethod
    def get_base(database_name: str) -> List[str]:
        database = Persistence.get_instance()._databases.get(database_name)
        if not database:
            raise IndexError(f"Database name ({database_name}) not exists.")

        if not hasattr(database, "get_base"):
            raise IndexError(f"Database ({database_name}) has not get_base method. ")

        return database.get_base()

    @staticmethod
    def get_databases() -> List[Database]:
        return list(Persistence.get_instance()._databases.values())

    @staticmethod
    def get_available_databases() -> List[str]:
        return list(Persistence.get_instance()._databases.keys())

    @staticmethod
    def get_available_models_for_database(database_name: str) -> List[str]:
        database = Persistence.get_instance()._databases.get(database_name)
        if not database:
            raise IndexError(f"Database name ({database_name}) not exists.")

        return list(database.get_model_names())

    @staticmethod
    def get_model(database_name: str, model_name: str):
        database = Persistence.get_instance()._databases.get(database_name)
        if not database:
            raise IndexError(f"Database name ({database_name}) not exists.")
        return database.get_model(model_name)

    @staticmethod
    def get_session(database_name: str):
        database = Persistence.get_instance()._databases.get(database_name)
        if not database:
            raise IndexError(f"Database name ({database_name}) not exists.")

        if not hasattr(database, "get_session"):
            raise IndexError(f"Database ({database_name}) has not get_session method. ")

        return database.get_session()

    @staticmethod
    def get_session_scope(database_name: str) -> Callable:
        database = Persistence.get_instance()._databases.get(database_name)
        if not database:
            raise IndexError(f"Database name ({database_name}) not exists.")

        if not hasattr(database, "get_session_scope"):
            raise IndexError(
                f"Database ({database_name}) has not get_session_scope method. "
            )

        return database.get_session_scope()



# from pika import BasicProperties
# from pika.exceptions import ChannelClosedByBroker
#
# from petisco.base.domain.message.domain_event import DomainEvent
# from petisco.base.domain.message.message_bus import MessageBus
# from petisco.extra.rabbitmq import RabbitMqMessageConfigurer
# from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_consumer_connector import RabbitMqConsumerConnector
# from petisco.extra.rabbitmq.application.message.formatter.rabbitmq_message_queue_name_formatter import \
#     RabbitMqMessageQueueNameFormatter
#
#
#
# class RabbitMqConsumerDomainEventBus(MessageBus):
#     def __init__(
#         self, connector: RabbitMqConsumerConnector, organization: str, service: str
#     ):
#         self.connector = connector
#         self.exchange_name = f"{organization}.{service}"
#         self.rabbitmq_key = f"publisher-{self.exchange_name}"
#         self.configurer = RabbitMqMessageConfigurer(connector, organization, service)
#         self.properties = BasicProperties(delivery_mode=2)  # PERSISTENT_TEXT_PLAIN
#
#     def publish(self, domain_event: DomainEvent):
#         if hasattr(self, "info_id"):
#             event = event.add_info_id(self.info_id)
#
#         if not event or not issubclass(event.__class__, Event):
#             raise TypeError("Bus only publishes petisco.Event objects")
#
#         try:
#             channel = self.connector.get_channel(self.rabbitmq_key)
#             routing_key = RabbitMqMessageQueueNameFormatter.format(
#                 event, exchange_name=self.exchange_name
#             )
#
#             channel.basic_publish(
#                 exchange=self.exchange_name,
#                 routing_key=routing_key,
#                 body=event.to_json(),
#                 properties=self.properties,
#             )
#         except ChannelClosedByBroker:
#             # If Event queue is not configured, it will be configured and publication retried.
#             self.configurer.configure_event(event)
#             self.publish(event)
#
#     def retry_publish_only_on_store_queue(self, event: Event):
#         if not event or not issubclass(event.__class__, Event):
#             raise TypeError("Bus only publishes petisco.Event objects")
#         channel = self.connector.get_channel(self.rabbitmq_key)
#         channel.basic_publish(
#             exchange=self.exchange_name,
#             routing_key="retry.store",
#             body=event.to_json(),
#             properties=self.properties,
#         )
#
#     def close(self):
#         self.connector.close(self.rabbitmq_key)



from abc import ABC
from typing import List

from pydantic import BaseModel, NonNegativeInt, PrivateAttr, validator

from petisco.base.domain.message.domain_event import DomainEvent
from petisco.base.domain.model.uuid import Uuid

DEFAULT_VERSION = 1


class AggregateRoot(ABC, BaseModel):
    aggregate_id: Uuid = None
    aggregate_version: NonNegativeInt = None
    _domain_events = PrivateAttr(default=[])

    @validator("aggregate_id", pre=True, always=True)
    def set_aggregate_id(cls, v):
        v = Uuid(v) if isinstance(v, str) else v
        return v or Uuid.v4()

    @validator("aggregate_version", pre=True, always=True)
    def set_aggregate_version(cls, v):
        return v or DEFAULT_VERSION

    def record(self, domain_event: DomainEvent):
        self._domain_events.append(domain_event)

    def clear_domain_events(self):
        self._domain_events = []

    def pull_domain_events(self) -> List[DomainEvent]:
        return self._domain_events

    def pull_first_domain_event(self) -> DomainEvent:
        return self._domain_events[0] if len(self._domain_events) > 0 else None

    def pull_last_domain_event(self) -> DomainEvent:
        return self._domain_events[-1] if len(self._domain_events) > 0 else None



from typing import List

from petisco.base.application.dependency_injection.dependency import Dependency
from petisco.base.domain.message.not_implemented_command_bus import (
    NotImplementedCommandBus,
)
from petisco.base.domain.message.not_implemented_domain_event_bus import (
    NotImplementedDomainEventBus,
)
from petisco.base.domain.message.not_implemented_message_comsumer import (
    NotImplementedMessageConsumer,
)
from petisco.base.domain.message.not_implemented_message_configurer import (
    NotImplementedMessageConfigurer,
)
from petisco.base.misc.builder import Builder


def get_default_message_dependencies() -> List[Dependency]:

    return [
        Dependency(
            name="domain_event_bus",
            default_builder=Builder(NotImplementedDomainEventBus),
            envar_modifier="PETISCO_DOMAIN_EVENT_BUS_TYPE",
        ),
        Dependency(
            name="command_bus",
            default_builder=Builder(NotImplementedCommandBus),
            envar_modifier="PETISCO_COMMAND_BUS_TYPE",
        ),
        Dependency(
            name="message_configurer",
            default_builder=Builder(NotImplementedMessageConfigurer),
            envar_modifier="PETISCO_MESSAGE_CONFIGURER_TYPE",
        ),
        Dependency(
            name="message_consumer",
            default_builder=Builder(NotImplementedMessageConsumer),
            envar_modifier="PETISCO_MESSAGE_CONSUMER_TYPE",
        ),
    ]


def get_rabbitmq_message_dependencies(
    organization: str, service: str, max_retries: int = 5
) -> List[Dependency]:
    from petisco.extra.rabbitmq.application.message.bus.rabbitmq_command_bus import (
        RabbitMqCommandBus,
    )
    from petisco.extra.rabbitmq.application.message.bus.rabbitmq_domain_event_bus import (
        RabbitMqDomainEventBus,
    )
    from petisco.extra.rabbitmq.application.message.configurer.rabbitmq_message_configurer import (
        RabbitMqMessageConfigurer,
    )
    from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_message_consumer import (
        RabbitMqMessageConsumer,
    )

    return [
        Dependency(
            name="domain_event_bus",
            default_builder=Builder(
                RabbitMqDomainEventBus, organization=organization, service=service
            ),
            envar_modifier="PETISCO_DOMAIN_EVENT_BUS_TYPE",
            builders={"not_implemented": Builder(NotImplementedDomainEventBus)},
        ),
        Dependency(
            name="command_bus",
            default_builder=Builder(
                RabbitMqCommandBus, organization=organization, service=service
            ),
            envar_modifier="PETISCO_COMMAND_BUS_TYPE",
            builders={"not_implemented": Builder(NotImplementedCommandBus)},
        ),
        Dependency(
            name="message_configurer",
            default_builder=Builder(
                RabbitMqMessageConfigurer,
                organization=organization,
                service=service,
            ),
            envar_modifier="PETISCO_MESSAGE_CONFIGURER_TYPE",
            builders={"not_implemented": Builder(NotImplementedMessageConfigurer)},
        ),
        Dependency(
            name="message_consumer",
            default_builder=Builder(
                RabbitMqMessageConsumer,
                organization=organization,
                service=service,
                max_retries=max_retries,
            ),
            envar_modifier="PETISCO_MESSAGE_CONSUMER_TYPE",
            builders={"not_implemented": Builder(NotImplementedMessageConsumer)},
        ),
    ]



import os

from setuptools import find_packages, setup

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
PACKAGE_NAME = "petisco"
VERSION = open("petisco/VERSION", "r").read().rstrip()

# The text of the README file
with open(os.path.join(CURRENT_DIR, "README.md")) as fid:
    README = fid.read()

with open("requirements/requirements.txt") as f:
    required = f.read().splitlines()

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description="Petisco is a framework for helping Python developers to build clean Applications",
    long_description=README,
    long_description_content_type="text/markdown",
    keywords=["DDD", "Use Case", "Clean Architecture", "REST", "Applications"],
    url="https://github.com/alice-biometrics/petisco",
    author="Alice Biometrics",
    author_email="support@alicebiometrics.com",
    license="MIT",
    install_requires=required,
    entry_points={
        "console_scripts": [
            "petisco = petisco.cli.petisco:main",
            "petisco-rabbitmq = petisco.cli.petisco_rabbitmq:main",
        ]
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    extras_require={
        "sqlalchemy": [
            "sqlalchemy==1.4.39",
            "sqlalchemy_utils==0.38.2",
            "PyMySQL==1.0.2",
        ],
        "redis": ["redis==4.1.1"],
        "rabbitmq": ["pika==1.2.1"],
        "slack": ["slack_sdk==3.15.2"],
        "elastic": ["elasticsearch<8.0.0,>=7.13.1"],
        "fastapi": ["fastapi==0.78.0"],
    },
)



from abc import abstractmethod

from petisco.base.application.notifier.notifier_exception_message import (
    NotifierExceptionMessage,
)
from petisco.base.application.notifier.notifier_message import NotifierMessage
from petisco.base.misc.interface import Interface


class Notifier(Interface):
    @abstractmethod
    def publish(self, notifier_message: NotifierMessage):
        raise NotImplementedError

    @abstractmethod
    def publish_exception(self, notifier_exception_message: NotifierExceptionMessage):
        raise NotImplementedError



from pika.adapters.blocking_connection import BlockingChannel


class RabbitMqConsumerConnector:
    def __init__(self, channel: BlockingChannel):
        self.channel = channel

    def get_connection(self, key_connection: str):
        raise RuntimeError("RabbitMqConsumerConnector works only with given channel.")

    def close(self, key_connection: str):
        pass

    def get_channel(self, key_connection: str) -> BlockingChannel:
        return self.channel
